/* $XConsortium: Geometry.c /main/64 1996/09/28 16:45:54 rws $ CHECKEDOUT */

/***********************************************************
Copyright 1987, 1988 by Digital Equipment Corporation, Maynard, Massachusetts
Copyright 1993 by Sun Microsystems, Inc. Mountain View, CA.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its 
documentation for any purpose and without fee is hereby granted, 
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in 
supporting documentation, and that the names of Digital or Sun not be
used in advertising or publicity pertaining to distribution of the
software without specific, written prior permission.  

DIGITAL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL
DIGITAL BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION,
ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
SOFTWARE.

SUN DISCLAIMS ALL WARRANTIES WITH REGARD TO  THIS  SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FIT-
NESS FOR A PARTICULAR PURPOSE. IN NO EVENT SHALL SUN BE  LI-
ABLE  FOR  ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE,  DATA  OR
PROFITS,  WHETHER  IN  AN  ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION  WITH
THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

/*

Copyright (c) 1987, 1988, 1994  X Consortium

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
X CONSORTIUM BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Except as contained in this notice, the name of the X Consortium shall not be
used in advertising or otherwise to promote the sale, use or other dealings
in this Software without prior written authorization from the X Consortium.

*/

//#include "IntrinsicI.h"
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/Shell.h>
#include <X11/ShellP.h>
#include <X11/StringDefs.h>
#include <X11/Intrinsic.h>

#define LOCK_PROCESS
#define UNLOCK_PROCESS
#define LOCK_APP(app)
#define UNLOCK_APP(app)
#define INIT_APP_LOCK(app)
#define FREE_APP_LOCK(app)
#define WIDGET_TO_APPCON(w)
#define DPY_TO_APPCON(d)
#define CALLGEOTAT(f)
extern String XtCXtToolkitError;

typedef struct _HookObjRec *HookObject;
typedef struct _HookObjPart {
    /* resources */
    XtCallbackList createhook_callbacks;
    XtCallbackList changehook_callbacks;
    XtCallbackList confighook_callbacks;
    XtCallbackList geometryhook_callbacks;
    XtCallbackList destroyhook_callbacks;
    WidgetList shells;
    Cardinal num_shells;
    /* private data */
    Cardinal max_shells;
    Screen* screen;
}HookObjPart;

typedef struct _HookObjRec {
    ObjectPart object;
    HookObjPart hooks;
} HookObjRec;

static void ClearRectObjAreas(r, old)
    RectObj r;
    XWindowChanges* old;
{
    Widget pw = _XtWindowedAncestor((Widget)r);
    int bw2;

    bw2 = old->border_width << 1;
    XClearArea( XtDisplay(pw), XtWindow(pw),
		old->x, old->y,
		old->width + bw2, old->height + bw2,
		TRUE );

    bw2 = r->rectangle.border_width << 1;
    XClearArea( XtDisplay(pw), XtWindow(pw),
		(int)r->rectangle.x, (int)r->rectangle.y,
		(unsigned int)(r->rectangle.width + bw2),
		(unsigned int)(r->rectangle.height + bw2),
		TRUE );
}

/*
 * Internal function used by XtMakeGeometryRequest and XtSetValues.
 * Returns more data than the public interface.  Does not convert
 * XtGeometryDone to XtGeometryYes.
 *
 * clear_rect_obj - *** RETURNED ***  
 *		    TRUE if the rect obj has been cleared, false otherwise.
 */

XtGeometryResult 
_XtMakeGeometryRequest (widget, request, reply, clear_rect_obj)
    Widget widget;
    XtWidgetGeometry *request, *reply;
    Boolean * clear_rect_obj;
{
    XtWidgetGeometry    junk;
    XtGeometryHandler manager = (XtGeometryHandler)0;
    XtGeometryResult returnCode;
    Widget parent = widget->core.parent;
    Boolean managed = False, parentRealized = False, rgm = False;
    XtConfigureHookDataRec req;
    Widget hookobj;

    *clear_rect_obj = FALSE;

    CALLGEOTAT(_XtGeoTrace(widget,
	"\"%s\" is making a %sgeometry request to its parent \"%s\".\n",
		   XtName(widget),
		   ((request->request_mode & XtCWQueryOnly))? "query only ":"",
		   (XtParent(widget))?XtName(XtParent(widget)):"Root"));
    CALLGEOTAT(_XtGeoTab(1));

    if (XtIsShell(widget)) {
	ShellClassExtension ext;
	LOCK_PROCESS;
	for (ext = (ShellClassExtension)((ShellWidgetClass)XtClass(widget))
		   ->shell_class.extension;
	     ext != NULL && ext->record_type != NULLQUARK;
	     ext = (ShellClassExtension)ext->next_extension);

	if (ext != NULL) {
	    if (  ext->version == XtShellExtensionVersion
		  && ext->record_size == sizeof(ShellClassExtensionRec)) {
		manager = ext->root_geometry_manager;
		rgm = True;
	    } else {
		String params[1];
		Cardinal num_params = 1;
		params[0] = XtClass(widget)->core_class.class_name;
		XtAppErrorMsg(XtWidgetToApplicationContext(widget),
		     "invalidExtension", "xtMakeGeometryRequest",
		     XtCXtToolkitError,
		     "widget class %s has invalid ShellClassExtension record",
		     params, &num_params);
	    }
	} else {
	    XtAppErrorMsg(XtWidgetToApplicationContext(widget),
			  "internalError", "xtMakeGeometryRequest",
			  XtCXtToolkitError,
			  "internal error; ShellClassExtension is NULL",
			  NULL, NULL);
	}
	managed = True;
	parentRealized = TRUE;
	UNLOCK_PROCESS;
    } else if (parent == NULL) {
	XtAppErrorMsg(XtWidgetToApplicationContext(widget),
		      "invalidParent","xtMakeGeometryRequest",XtCXtToolkitError,
		      "non-shell has no parent in XtMakeGeometryRequest",
		      (String *)NULL, (Cardinal *)NULL);
    } else /* not shell */ {
	managed = XtIsManaged(widget);

	if (XtIsComposite(parent)) {
	    parentRealized = XtIsRealized(parent);
	    LOCK_PROCESS;
	    manager = ((CompositeWidgetClass) (parent->core.widget_class))
		    ->composite_class.geometry_manager;
	    UNLOCK_PROCESS;
	} else if (managed) {
	    /* Should never happen - XtManageChildren should have checked */
	    XtAppErrorMsg(XtWidgetToApplicationContext(widget),
			  "invalidParent", "xtMakeGeometryRequest",
			  XtCXtToolkitError,
			  "XtMakeGeometryRequest - parent not composite",
			  (String *)NULL, (Cardinal *)NULL);
	} else {
	    /* no need to waste time checking if parent is actually realized
	     * at this point; since the child is unmanaged we need to perform
	     * the configure iff the child is realized, so we dummy the
	     * parentRealized checks below.
	     */
	    parentRealized = TRUE;
	}
    }

    if (managed && manager == (XtGeometryHandler) NULL) {
	XtErrorMsg("invalidGeometryManager","xtMakeGeometryRequest",
                 XtCXtToolkitError,
                 "XtMakeGeometryRequest - parent has no geometry manager",
                  (String *)NULL, (Cardinal *)NULL);
    }

    if (widget->core.being_destroyed) {
	CALLGEOTAT(_XtGeoTab(-1));
	CALLGEOTAT(_XtGeoTrace(widget,
		       "It is being destroyed, just return XtGeometryNo.\n"));
	return XtGeometryNo;
    }

    /* see if requesting anything to change */
    req.changeMask = 0;
    if (request->request_mode & CWStackMode
	&& request->stack_mode != XtSMDontChange) {
	    req.changeMask |= CWStackMode;
	    CALLGEOTAT(_XtGeoTrace(widget,
				   "Asking for a change in StackMode!\n"));
	    if (request->request_mode & CWSibling) {
		XtCheckSubclass(request->sibling, rectObjClass,
				"XtMakeGeometryRequest");
		req.changeMask |= CWSibling;
	    }
    }
    if (request->request_mode & CWX
	&& widget->core.x != request->x) {
	CALLGEOTAT(_XtGeoTrace(widget,
			       "Asking for a change in x: from %d to %d.\n",
			       widget->core.x, request->x));
	req.changeMask |= CWX;
    }
    if (request->request_mode & CWY
	&& widget->core.y != request->y) {
	CALLGEOTAT(_XtGeoTrace(widget,
			       "Asking for a change in y: from %d to %d.\n",
			       widget->core.y, request->y));
	req.changeMask |= CWY;
    }
    if (request->request_mode & CWWidth
	&& widget->core.width != request->width) {
	CALLGEOTAT(_XtGeoTrace(widget,"Asking for a change in width: from %d to %d.\n",
		       widget->core.width, request->width));
	req.changeMask |= CWWidth;
    }
    if (request->request_mode & CWHeight
	&& widget->core.height != request->height) {
	CALLGEOTAT(_XtGeoTrace(widget,
			    "Asking for a change in height: from %d to %d.\n",
			       widget->core.height, request->height));
	req.changeMask |= CWHeight;
    }
    if (request->request_mode & CWBorderWidth
	&& widget->core.border_width != request->border_width){
	CALLGEOTAT(_XtGeoTrace(widget,
		       "Asking for a change in border_width: from %d to %d.\n",
		       widget->core.border_width, request->border_width));
	req.changeMask |= CWBorderWidth;
    }
    if (! req.changeMask) {
	CALLGEOTAT(_XtGeoTrace(widget,
		       "Asking for nothing new,\n"));
	CALLGEOTAT(_XtGeoTab(-1));
	CALLGEOTAT(_XtGeoTrace(widget,
		       "just return XtGeometryYes.\n"));
	return XtGeometryYes;
    }
    req.changeMask |= (request->request_mode & XtCWQueryOnly);

    if ( !(req.changeMask & XtCWQueryOnly) && XtIsRealized(widget) ) {
	/* keep record of the current geometry so we know what's changed */
	req.changes.x = widget->core.x ;
	req.changes.y = widget->core.y ;
	req.changes.width = widget->core.width ;
	req.changes.height = widget->core.height ;
	req.changes.border_width = widget->core.border_width ;
    }

    if (!managed || !parentRealized) {
	CALLGEOTAT(_XtGeoTrace(widget,
			       "Not Managed or Parent not realized.\n"));
	/* Don't get parent's manager involved--assume the answer is yes */
	if (req.changeMask & XtCWQueryOnly) {
	    /* He was just asking, don't change anything, just tell him yes */
	    CALLGEOTAT(_XtGeoTrace(widget,"QueryOnly request\n"));
	    CALLGEOTAT(_XtGeoTab(-1));
	    CALLGEOTAT(_XtGeoTrace(widget,"just return XtGeometryYes.\n"));
	    return XtGeometryYes;
	} else {
	    CALLGEOTAT(_XtGeoTrace(widget,
				   "Copy values from request to widget.\n"));
	    /* copy values from request to widget */
	    if (request->request_mode & CWX)
		widget->core.x = request->x;
	    if (request->request_mode & CWY)
		widget->core.y = request->y;
	    if (request->request_mode & CWWidth)
		widget->core.width = request->width;
	    if (request->request_mode & CWHeight)
		widget->core.height = request->height;
	    if (request->request_mode & CWBorderWidth)
		widget->core.border_width = request->border_width;
	    if (!parentRealized) {
		CALLGEOTAT(_XtGeoTab(-1));
		CALLGEOTAT(_XtGeoTrace(widget,"and return XtGeometryYes.\n"));
		return XtGeometryYes;
	    }
	    else returnCode = XtGeometryYes;
	}
    } else {
	/* go ask the widget's geometry manager */
	CALLGEOTAT(_XtGeoTrace(widget,
			       "Go ask the parent geometry manager.\n"));
	if (reply == (XtWidgetGeometry *) NULL) {
	    returnCode = (*manager)(widget, request, &junk);
	} else {
	    returnCode = (*manager)(widget, request, reply);
	}
    }

    /*
     * If Unrealized, not a XtGeometryYes, or a query-only then we are done.
     */

    if ((returnCode != XtGeometryYes) || 
	(req.changeMask & XtCWQueryOnly) || !XtIsRealized(widget)) {

#ifdef XT_GEO_TATTLER
	switch(returnCode){
	case XtGeometryNo:
	    CALLGEOTAT(_XtGeoTab(-1));
	    CALLGEOTAT(_XtGeoTrace(widget,"\"%s\" returns XtGeometryNo.\n",
			  (XtParent(widget))?XtName(XtParent(widget)):"Root"));
            /* check for no change */
            break ;
	case XtGeometryDone:
	    CALLGEOTAT(_XtGeoTab(-1));
	    CALLGEOTAT(_XtGeoTrace(widget,"\"%s\" returns XtGeometryDone.\n",
		        (XtParent(widget))?XtName(XtParent(widget)):"Root"));
            /* check for no change in queryonly */
            break ;
	case XtGeometryAlmost:
	    CALLGEOTAT(_XtGeoTab(-1));
	    CALLGEOTAT(_XtGeoTrace(widget,"\"%s\" returns XtGeometryAlmost.\n",
			 (XtParent(widget))?XtName(XtParent(widget)):"Root"));
	    CALLGEOTAT(_XtGeoTab(1));
	    CALLGEOTAT(_XtGeoTrace(widget,"Proposal: width %d height %d.\n",
			   (reply)?reply->width:junk.width,
			   (reply)?reply->height:junk.height));
	    CALLGEOTAT(_XtGeoTab(-1));
	    
            /* check for no change */
            break ;
	case XtGeometryYes:
	    if (req.changeMask & XtCWQueryOnly) {
		CALLGEOTAT(_XtGeoTrace(widget,
				"QueryOnly specified, no configuration.\n"));
	    }
            if (!XtIsRealized(widget)) {
		CALLGEOTAT(_XtGeoTrace(widget,
				 "\"%s\" not realized, no configuration.\n",
				 XtName(widget)));
	    }
	    CALLGEOTAT(_XtGeoTab(-1));
            CALLGEOTAT(_XtGeoTrace(widget,"\"%s\" returns XtGeometryYes.\n",
                         (XtParent(widget))?XtName(XtParent(widget)):"Root"));
	    break ;
	}
#endif
	return returnCode;
    }

    CALLGEOTAT(_XtGeoTab(-1));
    CALLGEOTAT(_XtGeoTrace(widget,"\"%s\" returns XtGeometryYes.\n",
		   (XtParent(widget))?XtName(XtParent(widget)):"Root"));

    if (XtIsWidget(widget)) {	/* reconfigure the window (if needed) */

	if (rgm) return returnCode;

	if (req.changes.x != widget->core.x) {
 	    req.changeMask |= CWX;
 	    req.changes.x = widget->core.x;
	    CALLGEOTAT(_XtGeoTrace(widget,
				   "x changing to %d\n",widget->core.x));
 	}
 	if (req.changes.y != widget->core.y) {
 	    req.changeMask |= CWY;
 	    req.changes.y = widget->core.y;
	    CALLGEOTAT(_XtGeoTrace(widget,
				   "y changing to %d\n",widget->core.y));
 	}
 	if (req.changes.width != widget->core.width) {
 	    req.changeMask |= CWWidth;
 	    req.changes.width = widget->core.width;
	    CALLGEOTAT(_XtGeoTrace(widget,
				 "width changing to %d\n",widget->core.width));
 	}
 	if (req.changes.height != widget->core.height) {
 	    req.changeMask |= CWHeight;
 	    req.changes.height = widget->core.height;
	    CALLGEOTAT(_XtGeoTrace(widget,
                               "height changing to %d\n",widget->core.height));
 	}
 	if (req.changes.border_width != widget->core.border_width) {
 	    req.changeMask |= CWBorderWidth;
 	    req.changes.border_width = widget->core.border_width;
	    CALLGEOTAT(_XtGeoTrace(widget,
				   "border_width changing to %d\n",
				   widget->core.border_width));
 	}
	if (req.changeMask & CWStackMode) {
	    req.changes.stack_mode = request->stack_mode;
	    CALLGEOTAT(_XtGeoTrace(widget,"stack_mode changing\n"));
	    if (req.changeMask & CWSibling) {
		if (XtIsWidget(request->sibling))
		    req.changes.sibling = XtWindow(request->sibling);
		else
		    req.changeMask &= ~(CWStackMode | CWSibling);
	    }
	}

#ifdef XT_GEO_TATTLER
        if (req.changeMask) {
	    CALLGEOTAT(_XtGeoTrace(widget,
				   "XConfigure \"%s\"'s window.\n",
				   XtName(widget)));
	} else {
	    CALLGEOTAT(_XtGeoTrace(widget,
			     "No window configuration needed for \"%s\".\n",
			     XtName(widget)));
	}
#endif

	XConfigureWindow(XtDisplay(widget), XtWindow(widget),
			 req.changeMask, &req.changes);
    }
    else {			/* RectObj child of realized Widget */
	*clear_rect_obj = TRUE;
	CALLGEOTAT(_XtGeoTrace(widget,
			       "ClearRectObj on \"%s\".\n",XtName(widget)));

	ClearRectObjAreas((RectObj)widget, &req.changes);
    }
    hookobj = XtHooksOfDisplay(XtDisplayOfObject(widget));;
    if (XtHasCallbacks(hookobj, XtNconfigureHook) == XtCallbackHasSome) {
	req.type = XtHconfigure;
	req.widget = widget;
	XtCallCallbackList(hookobj,
		((HookObject)hookobj)->hooks.confighook_callbacks,
		(XtPointer)&req);
    }

    return returnCode;
} /* _XtMakeGeometryRequest */
